package br.com.estruturadedados.utilitario.lista.vetor;

import br.com.estruturadedados.utilitario.colecao.Iterator;
import br.com.estruturadedados.utilitario.lista.Lista;

public class ListaVetor<E extends Comparable<E>> implements Lista<E> {

    private int capacidade;
    private E[] elementos;
    private int tamanho;

    public ListaVetor() {
        this.capacidade = 10;
        this.elementos = (E[]) new Integer[this.capacidade];
        this.tamanho = 0;
    }

    public ListaVetor(int capacidadeInicial) throws Exception {
        this.validarCapacidadeInicial(capacidadeInicial);
        if (capacidadeInicial > 0) {
            this.capacidade = capacidadeInicial;
        } else {
            if (this.capacidade == 0) {
            }
        }
        switch (capacidadeInicial) {
            case 0:
                this.capacidade = 10;
                break;
        }
        this.elementos = (E[]) new Object[this.capacidade];
        this.tamanho = 0;
    }

    @Override
    public void adiciona(E elemento, int posicao) throws Exception {
        this.verificarSeEstaCeheio();
        this.validarPosicao(posicao);
        this.verificarPosicaoOcupada(posicao);
        this.elementos[posicao] = elemento;
        this.tamanho += 1;
    }

    @Override
    public void adicionaInicio(E elemento) {
        this.verificarSeEstaCeheio();
        E[] novosElementos = (E[]) new Integer[this.capacidade];
        novosElementos[0] = elemento;
        this.transferirElementos(novosElementos, 0, 1);
        this.tamanho += 1;
    }

    @Override
    public void adicionaFim(E elemento) {
        this.verificarSeEstaCeheio();
        this.elementos[this.tamanho] = elemento;
        this.tamanho += 1;
    }

    @Override
    public E recupera(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        E elemento = this.elementos[posicao];
        return elemento;
    }

    @Override
    public boolean vazio() {
        return this.tamanho == 0;
    }

    @Override
    public void remove(int posicao) throws Exception {
        this.validarPosicao(posicao);
        this.verificarSeEstaVazio();
        this.verificarPosicaoLivre(posicao);
        E elementoRemovido = this.elementos[posicao];
        E elementoNulo = null;
        this.elementos[posicao] = this.elementos[posicao + 1];
        this.elementos[posicao + 1] = elementoNulo;
        this.tamanho -= 1;
    }

    @Override
    public void removeInicio() throws Exception {
        int posicao = 0;
        this.verificarSeEstaVazio();
        E elementoRemovido = this.elementos[posicao];
        E[] novosElementos = (E[]) new Object[this.capacidade];
        this.transferirElementos(novosElementos, 1, 0);
        this.tamanho -= 1;
    }

    @Override
    public void removeFim() throws Exception {
        int posicao = this.tamanho - 1;
        this.verificarSeEstaVazio();
        E elemento = this.elementos[posicao];
        E elementoNulo = null;
        this.elementos[posicao] = elementoNulo;
        this.tamanho -= 1;
    }

    @Override
    public int tamanho() {
        return this.tamanho;
    }

    @Override
    public void limpa() throws Exception {
        this.verificarSeEstaVazio();
        this.elementos = null;
    }

    @Override
    public Iterator<E> iterator() {
        Iterator<E> iterator = new Iterador<>(this.elementos);
        return iterator;
    }

    public E[] getElementos() {
        return elementos;
    }

    public void setElementos(E[] elementos) {
        this.elementos = elementos;
    }

    private void verificarSeEstaCeheio() {
        if (this.tamanho >= this.capacidade) {
            this.redimensiona();
        }
    }

    private void redimensiona() {
        E[] vetorRedimensionado = (E[]) new Integer[this.capacidade * 2];
        System.arraycopy(this.elementos, 0, vetorRedimensionado, 0, this.tamanho);
        this.elementos = vetorRedimensionado;
        this.capacidade = vetorRedimensionado.length;
    }

    private void validarPosicao(int posicao) throws Exception {
        if (posicao < 0 || posicao >= this.capacidade) {
            Exception exception = new Exception("Posição inválida.");
            throw exception;
        }
    }

    private void verificarPosicaoOcupada(int posicao) throws Exception {
        if (this.elementos[posicao] != null) {
            Exception exception = new Exception("Posição ocupada.");
            throw exception;
        }
    }

    private void verificarPosicaoLivre(int posicao) throws Exception {
        if (this.elementos[posicao] == null) {
            Exception exception = new Exception("Posição desoculpada.");
            throw exception;
        }
    }

    private void verificarSeEstaVazio() throws Exception {
        if (this.vazio()) {
            Exception exception = new Exception("Está vazio.");
            throw exception;
        }
    }

    private void validarCapacidadeInicial(int capacidadeInicial) throws Exception {
        if (capacidadeInicial < 0) {
            Exception exception = new Exception("Capacidade inválida: " + capacidadeInicial);
            throw exception;
        }
    }

    private void transferirElementos(E[] novosElementos, int posicaoInicialVetor1, int posicaoInicialVetor2) {
        System.arraycopy(this.elementos, posicaoInicialVetor1, novosElementos, posicaoInicialVetor2, this.capacidade - 1);
        this.elementos = novosElementos;
    }
}
